#import grok
import grokcore.component
from zope.schema.interfaces import IField
from zope.interface import Interface
from zope.interface.interface import InterfaceClass
from zope.interface.interfaces import IInterface
from zope.component import getUtility
from sqlalchemy.types import String, Integer
from megrok.rdb import Model
from zope.schema import Int
from zope.schema import Text
from zope.schema import TextLine

#def Fields(model):
#    return grok.Fields(schema_from_model(model))

def schema_from_model(model):
    table = model.__table__
    bases = (Interface,)
    attrs = {}
    for i, column in enumerate(table.columns):
        if len(column.foreign_keys) or column.primary_key:
            continue
        field = IField(column.type)
        field.__name__ = str(column.name)
        field.title = unicode(column.name)
        field.required = not column.nullable
        attrs[column.name] = field
        field.order = i
       
    return InterfaceClass(name=model.__table__.name,
                          bases=bases,
                          attrs=attrs,
                          __doc__='Generated from metadata')

@grokcore.component.adapter(String)
@grokcore.component.implementer(IField)
def field_from_sa_string(s):
    return TextLine(__name__ = u'__dummy__',
                title = u'__dummy__')

@grokcore.component.adapter(Integer)
@grokcore.component.implementer(IField)
def field_from_sa_integer(i):
    return Int(__name__ = u'__dummy__',
                title = u'__dummy__')




from sqlalchemy.orm.collections import MappedCollection, collection
from sqlalchemy.orm.util import identity_key

from zope.interface import implements
from zope.location.interfaces import ILocation

from grokcore.component import Context
from megrok.rdb.interfaces import IContainer

from megrok.rdb import directive
from z3c.saconfig import Session

class Model(Context):
    implements(ILocation)

    __parent__ = None
    __name__ = None
    
    def __init__(self, **kwargs):
        # XXX can we use the __init__ that sqlalchemy.ext.declarative sets up?
        for k in kwargs:
            if not hasattr(type(self), k):
                raise TypeError('%r is an invalid keyword argument for %s' %
                                (k, type(self).__name__))
            setattr(self, k, kwargs[k])

class LocatedModel(Model):
    implements(ILocation)

    @property
    def __parent__(self):
        return

    @property
    def __name__(self):
        return
    
def default_keyfunc(node):
    primary_keys = node.__table__.primary_key.keys()
    if len(primary_keys) == 1:
        return getattr(node, primary_keys[0])
    else:
        raise RuntimeError(
            "don't know how to do keying with composite primary keys")

class Container(MappedCollection):
    implements(IContainer, ILocation)

    __parent__ = None
    __name__ = None
    
    def __init__(self, *args, **kw):
        rdb_key = directive.key.bind().get(self)
        if rdb_key:
            keyfunc = lambda node:getattr(node, rdb_key)
        elif hasattr(self, 'keyfunc'):
            keyfunc = self.keyfunc
        else:
            keyfunc = default_keyfunc
        MappedCollection.__init__(self, keyfunc=keyfunc)

    @collection.on_link
    def on_link(self, adapter):
        if adapter is not None:
            self.__parent__ = adapter.owner_state.obj()
            self.__name__ = unicode(adapter.attr.key)
        else:
            # unlinking collection from parent
            self.__parent__ = None
            self.__name__ = None
            
    def __setitem__(self, key, item):
        key = unicode(key)
        self._receive(item, key)
        MappedCollection.__setitem__(self, key, item)

    def __delitem__(self, key):
        key = unicode(key)
        self._release(self[key])
        MappedCollection.__delitem__(self, key)
        
    def _receive(self, item, key):
        item.__name__ = key
        item.__parent__ = self

    def _release(self, item):
        del item.__name__
        del item.__parent__
    
    @collection.internally_instrumented
    @collection.appender
    def set(self, value, _sa_initiator=None):
        key = self.keyfunc(value)
        if key is None:
            session = Session()
            session.flush()
            key = self.keyfunc(value)
        self.__setitem__(key, value, _sa_initiator)

class QueryContainer(object):
    """A read-only Container backed by set of SQLAlchemy queries.

    This container is often used as the "root" object for an application.
    """
    # XXX should really only implement IReadContainer, but this requires
    # a bit of refactoring in Grok
    implements(IContainer, ILocation)

    def query(self):
        """Implement this returning the result of a query.
        """
        raise NotImplementedError

    def convert(self, value):
        """Override this to convert output of query container.
        """
        return value

    def __getitem__(self, key):
        result = self.dbget(key)
        if result is None:
            raise KeyError(key)
        result = self.convert(result)
        result.__parent__ = self
        result.__name__ = key
        return result

    def dbget(self, key):        
        return self.query().get(key)

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default
    
    def __contains__(self, key):
        try:
            self[key]
            return True
        except KeyError:
            return False

    def has_key(self, key):
        return key in self

    def keys(self):
        result = []
        # XXX probably not the most efficient way to do this
        for v in self.query().all():
            result.append(self.keyfunc(v))
        return result

    def __iter__(self):
        for key in self.keys():
            yield key
    
    def values(self):
        result = []
        for v in self.query().all():
            converted = self.convert(v)
            converted.__parent__ = self
            converted.__name__ = self.keyfunc(v)
            result.append(converted)
        return result
        
    def items(self):
        result = []
        for v in self.query().all():
            converted = self.convert(v)
            converted.__parent__ = self
            key = converted.__name__ = self.keyfunc(v)
            result.append((key, converted))
        return result

    def __len__(self):
        return self.query().count()
    
    def keyfunc(self, value):
        """Get the key for this value.
        """
        # XXX this isn't very general...
        class_, key_tuple = identity_key(instance=value)
        return unicode(key_tuple[0])




"""
ORM-managed objects such as collection attributes and the items in them should
have an ILocation so it is easy to generate a URL for them.

Let's first grok things::

  >>> from grokcore.component.testing import grok
  >>> grok('megrok.rdb.meta')
  >>> grok(__name__)

We need to set up an engine::

  >>> from megrok.rdb.testing import configureEngine
  >>> engine = configureEngine()
  
We now need to reflect the tables in our database to our classes::

  >>> rdb.setupDatabase(metadata)

Let's start using the database now::

  >>> session = rdb.Session()
  >>> philosophy = Department(name='Philosophy')
  >>> session.add(philosophy)

An unconnected object has ``__name__`` and ``__parent__`` of None::

  >>> print philosophy.__name__
  None
  >>> print philosophy.__parent__
  None

The ``courses`` attribute of ``philosophy`` however does have a
``__name__`` and ``__parent__``::

  >>> philosophy.courses.__name__
  u'courses'
  >>> philosophy.courses.__parent__ is philosophy
  True

Let's create some objects to put in the ``courses`` attribute of
``philosophy``:

  >>> logic = Course(name='Logic')
  >>> ethics = Course(name='Ethics')
  >>> metaphysics = Course(name='Metaphysics')
  >>> session.add_all([logic, ethics, metaphysics])

At this stage these object are also unconnected, so have ``__name__``
and ``__parent__`` of None::

  >>> print logic.__name__
  None
  >>> print logic.__parent__
  None

Let's now add them to the courses container::

  >>> philosophy.courses.set(logic)
  >>> philosophy.courses.set(ethics)
  >>> philosophy.courses.set(metaphysics)

They now should have the proper parents::

  >>> philosophy.courses['1'].__name__
  u'1'
  >>> philosophy.courses['1'].__parent__ is philosophy.courses
  True
  
"""

from megrok import rdb

from sqlalchemy import Column, ForeignKey
from sqlalchemy.types import Integer, String
from sqlalchemy.orm import relation

metadata = rdb.MetaData()

rdb.metadata(metadata)
    
class Courses(rdb.Container):
    pass

class Department(rdb.Model):    
    id = Column('id', Integer, primary_key=True)
    name = Column('name', String(50))
    courses = relation('Course', 
                       backref='department',
                       collection_class=Courses)

class Course(rdb.Model):
    id = Column('id', Integer, primary_key=True)
    department_id = Column('department_id', Integer, 
                           ForeignKey('department.id'))
    name = Column('name', String(50))



import martian
from martian.error import GrokError
from sqlalchemy.ext.declarative import instrument_declarative
import sqlalchemy

from megrok import rdb

def default_tablename(factory, module, **data):
    return factory.__name__.lower()

class ModelGrokker(martian.ClassGrokker):
    martian.component(rdb.Model)
    martian.directive(rdb.tablename, get_default=default_tablename)
    martian.directive(rdb.metadata)
    martian.directive(rdb.reflected)
    martian.directive(rdb.table)
    martian.directive(rdb.tableargs)
    martian.directive(rdb.inherits)
    martian.directive(rdb.polymorphic_identity)
    martian.directive(rdb.polymorphic_on)
    
    def execute(self, class_, tablename, metadata, reflected, table,
                tableargs, inherits, polymorphic_identity, polymorphic_on,
                **kw):
        if table is not None:
            class_.__table__ = table
        else:
            class_.__tablename__ = tablename
        if tableargs is not None:
            class_.__table_args__ = tableargs
        
        if reflected:
            if not hasattr(metadata, '_reflected_registry'):
                metadata._reflected_registry = {}

            metadata._reflected_registry[class_] = dict(
                inherits=inherits,
                polymorphic_on=polymorphic_on,
                polymorphic_identity=polymorphic_identity,
                )


            # if this table is reflected, don't instrument now but
            # manually map later
            return True
        # we associate the _decl_registry with the metadata object
        # to make sure it's unique per metadata. A bit of a hack..
        if not hasattr(metadata, '_decl_registry'):
            metadata._decl_registry = {}
        try:
            instrument_declarative(class_, metadata._decl_registry, metadata)
        except sqlalchemy.exc.InvalidRequestError:
            # FIXME scary - catching too many errors
            # allows re-grokking of classes that were already
            # instrumented. Better would be to un-instrument classes
            # after tests, but how to uninstrument?
            pass

        return True
    
class ContainerGrokker(martian.ClassGrokker):
    martian.component(rdb.Container)
    
    def grok(self, name, factory, module_info, config, **kw):
        rdb_key = rdb.key.bind().get(factory)
        if rdb_key and hasattr(factory, 'keyfunc'):
            raise GrokError(
                "It is not allowed to specify a custom 'keyfunc' method "
                "for rdb.Container %r, when a rdb.key directive has also "
                "been given." % factory, factory)
        return True



"""
It's possible to directly specify a table using ``rdb.table`` as well.

Let's first grok things::

  >>> from grokcore.component.testing import grok
  >>> grok('megrok.rdb.meta')
  >>> grok(__name__)

We need to set up an engine::

  >>> from megrok.rdb.testing import configureEngine
  >>> engine = configureEngine()
  
We now need to create the tables we defined in our database::

  >>> rdb.setupDatabase(metadata)

Let's start using the database now::

  >>> session = rdb.Session()
  >>> dummy = User(name=u'dummy', password=u'secret', role=u'admin')
  >>> session.add(dummy)
  >>> silly = User(name=u'silly', password=u'secret', role=u'admin')
  >>> session.add(silly)

We can now verify that the users are there::

  >>> [(user.id, user.name, user.password, user.role) for user in
  ... session.query(User)]
  [(1, u'dummy', u'secret', u'admin'), (2, u'silly', u'secret', u'admin')]
    
"""


from megrok import rdb

from sqlalchemy import (Table, Column, MetaData, Integer, Unicode)

metadata = rdb.MetaData()


user_table = Table(
    'user_table', metadata,
    Column('id', Integer, primary_key=True),
    Column('name', Unicode(20), nullable=False, index=True, unique=True),
    Column('password', Unicode(100), nullable=False),
    Column('role', Unicode(20), nullable=False, index=True),
    )

class User(rdb.Model):
    rdb.metadata(metadata)
    rdb.table(user_table)



from zope import component

from z3c.saconfig import EngineFactory
from z3c.saconfig.interfaces import IEngineFactory

from z3c.saconfig import GloballyScopedSession
from z3c.saconfig.interfaces import IScopedSession

def configureEngine(TEST_DSN='sqlite:///:memory:'):
    """Utility function for tests to set up an in-memory test database.

    Returns engine object.
    """
    engine_factory = EngineFactory(TEST_DSN)
    component.provideUtility(engine_factory, provides=IEngineFactory)

    scoped_session = GloballyScopedSession()
    component.provideUtility(scoped_session, provides=IScopedSession)
    
    return engine_factory()




from martian import MarkerDirective, Directive, CLASS, CLASS_OR_MODULE, ONCE

# XXX add proper validation logic

class key(Directive):
    scope = CLASS
    store = ONCE
    default = u''

class metadata(Directive):
    scope = CLASS_OR_MODULE
    store = ONCE
    default = None

class tablename(Directive):
    scope = CLASS
    store = ONCE
    default = u''

class reflected(MarkerDirective):
    scope = CLASS_OR_MODULE
    store = ONCE

class inherits(Directive):
    scope = CLASS
    store = ONCE
    default = None

class polymorphic_on(Directive):
    scope = CLASS
    store = ONCE
    default = None
    
    def factory(self, table, column):
        return table, column

class polymorphic_identity(Directive):
    scope = CLASS
    store = ONCE
    default = None

class table(Directive):
    scope = CLASS
    store = ONCE
    default = None

class tableargs(Directive):
    scope = CLASS
    store = ONCE
    default = None

    def factory(self, *args, **kw):
        if args:
            return args + (kw,)
        elif kw:
            return kw
        else:
            return None



class locatedproperty(property):
    """A property that locates whatever is returned.

    This is useful to create properties that fulfill the ILocation protocol
    and therefore have a __parent__ and a __name__. Use it like this:

    @locatedproperty
    def whatever(self):
        return Something()

    now obj.whatever will have a __parent__ set to obj and a __name__
    set to 'whatever'.

    It's quite possible we'll find a nicer spelling for this
    eventually for the purposes of megrok.rdb, but this works for now.
    """
    
    def __init__(self, f):
        super(locatedproperty, self).__init__(f)
        self.__name__ = f.__name__

    def __get__(self, obj, type=None):
        value = super(locatedproperty, self).__get__(obj, type)
        value.__parent__ = obj
        value.__name__ = self.__name__
        return value



from zope.interface import Interface, Attribute, implements

from zope.app.container.interfaces import IContainer as IContainerBase
from grokcore.component.interfaces import IContext

# FIXME: this doesn't depend on grok.interfaces.IContainer. If something
# is registered directly for those containers, it won't show up for
# this IContainer... Want to avoid directly having to depend on Grok
# with megrok.rdb however
class IContainer(IContext, IContainerBase):
    def set(value):
        """Add a new value to the container without having to specify the key.

        Lets the container figure out an appropriate key.


        Defined by SQLAlchemy dictionary-based collections.
        """

    def remove(value):
        """Remove a value from the container, by value.

        
        Defined by SQLAlchemy dictionary-based collections.
        """

class IDatabaseSetupEvent(Interface):
    metadata = Attribute("Metadata that was just set up")

class DatabaseSetupEvent(object):
    implements(IDatabaseSetupEvent)
    
    def __init__(self, metadata):
        self.metadata = metadata



"""
A reflected model is a model that is defined by the table in the
database (as opposed to defining the database table in Python).

Let's first grok things::

  >>> from grokcore.component.testing import grok
  >>> grok('megrok.rdb.meta')
  >>> grok(__name__)

We need to set up an engine::

  >>> from megrok.rdb.testing import configureEngine
  >>> engine = configureEngine()

Let's set up the tables::

  >>> from sqlalchemy.sql import text
  >>> conn = engine.connect()
  >>> s = text('''
  ...   create table department (
  ...     id integer,
  ...     name char(50),
  ...     primary key (id))
  ... ''')
  >>> result = conn.execute(s)
  >>> s = text('''
  ...   create table course (
  ...     id integer,
  ...     name char(50),
  ...     department_id integer,
  ...     primary key (id),
  ...     foreign key (department_id) references department(id))
  ... ''')
  >>> result = conn.execute(s)
  
We now need to reflect the tables in our database to our classes::

  >>> rdb.setupDatabase(metadata)

Let's start using the database now::

  >>> session = rdb.Session()
  >>> philosophy = Department(name='Philosophy')
  >>> session.add(philosophy)
  >>> logic = Course(name='Logic')
  >>> ethics = Course(name='Ethics')
  >>> metaphysics = Course(name='Metaphysics')
  >>> session.add_all([logic, ethics, metaphysics])
  
Let's now add them to the courses container::

  >>> philosophy.courses.set(logic)
  >>> philosophy.courses.set(ethics)
  >>> philosophy.courses.set(metaphysics)

We can now verify that the courses are there::

  >>> [(course.id, course.name, course.department_id) for course in
  ... session.query(Course)]
  [(1, 'Logic', 1), (2, 'Ethics', 1), (3, 'Metaphysics', 1)]

  >>> for key, value in sorted(philosophy.courses.items()):
  ...   print key, value.name, value.department.name
  1 Logic Philosophy
  2 Ethics Philosophy
  3 Metaphysics Philosophy
"""


from megrok import rdb

from sqlalchemy import Column, ForeignKey
from sqlalchemy.types import Integer, String
from sqlalchemy.orm import relation

metadata = rdb.MetaData()

rdb.metadata(metadata)

class Courses(rdb.Container):
    pass

class Department(rdb.Model):
    rdb.reflected()
    courses = relation('Course', 
                       backref='department',
                       collection_class=Courses)

class Course(rdb.Model):
    rdb.reflected()



import unittest

from megrok import rdb

import grokcore.component.testing

from sqlalchemy import Column
from sqlalchemy.types import Integer
from sqlalchemy.schema import ForeignKeyConstraint

class TableArgs(unittest.TestCase):
    def setUp(self):
        grokcore.component.testing.grok('megrok.rdb.meta')

    def test_no_tableargs(self):
        class MyClass(rdb.Model):
            rdb.metadata(rdb.MetaData())
            id = Column(Integer, primary_key=True)
            
        grokcore.component.testing.grok_component('MyClass', MyClass)
        self.assert_(not hasattr(MyClass, '__table_args__'))
    
    def test_empty_tableargs(self):
        class MyClass(rdb.Model):
            rdb.metadata(rdb.MetaData())
            id = Column(Integer, primary_key=True)
            rdb.tableargs()
    
        grokcore.component.testing.grok_component('MyClass', MyClass)
        self.assert_(not hasattr(MyClass, '__table_args__'))

    def test_non_keyword_tableargs(self):
        class MyClass(rdb.Model):
            rdb.metadata(rdb.MetaData())
            rdb.tableargs(ForeignKeyConstraint(['id'], ['whah.id']))
            id = Column(Integer, primary_key=True)
            
        grokcore.component.testing.grok_component('MyClass', MyClass)
        arg, kw = MyClass.__table_args__
        self.assert_(isinstance(arg, ForeignKeyConstraint))
        self.assertEquals({}, kw)

    def test_keyword_tableargs(self):
        class MyClass(rdb.Model):
            rdb.metadata(rdb.MetaData())
            rdb.tableargs(schema='bar')
            id = Column(Integer, primary_key=True)
            
        grokcore.component.testing.grok_component('MyClass', MyClass)
        self.assertEquals({'schema': 'bar'}, MyClass.__table_args__)

    def test_both_tableargs(self):
        class MyClass(rdb.Model):
            rdb.metadata(rdb.MetaData())
            rdb.tableargs(ForeignKeyConstraint(['id'], ['whah.id']),
                          schema='bar')
            id = Column(Integer, primary_key=True)
            
        grokcore.component.testing.grok_component('MyClass', MyClass)
        arg = MyClass.__table_args__[:-1]
        kw = MyClass.__table_args__[-1]
        self.assertEquals(1, len(arg))
        self.assert_(isinstance(arg[0], ForeignKeyConstraint))
        self.assertEquals({'schema': 'bar'}, kw)
        
def suite():
    return unittest.makeSuite(TableArgs)




"""
A model by default also defines a relationa l database table in Python
(as opposed to reflecting it from the database schema).

Let's first grok things::

  >>> from grokcore.component.testing import grok
  >>> grok('megrok.rdb.meta')
  >>> grok(__name__)

We need to set up an engine::

  >>> from megrok.rdb.testing import configureEngine
  >>> engine = configureEngine()

When the database is set up there is an ``rdb.IDatabaseSetupEvent``
that can be hooked into to do some extra configuration. The event is
fired after all other database is completed. Here we just print
something so we can test whether it was fired correctly::

  >>> from zope import component
  >>> @component.adapter(rdb.IDatabaseSetupEvent)
  ... def setupHandler(event):
  ...    print 'Database setup event'
  >>> component.provideHandler(setupHandler)
  
We now need to create the tables we defined in our database::

  >>> rdb.setupDatabase(metadata)
  Database setup event

As we can see the database setup event was indeed fired.

Let's start using the database now::

  >>> session = rdb.Session()
  >>> philosophy = Department(name='Philosophy')
  >>> session.add(philosophy)
  >>> logic = Course(name='Logic')
  >>> ethics = Course(name='Ethics')
  >>> metaphysics = Course(name='Metaphysics')
  >>> session.add_all([logic, ethics, metaphysics])
  
Let's now add them to the courses container::

  >>> philosophy.courses.set(logic)
  >>> philosophy.courses.set(ethics)
  >>> philosophy.courses.set(metaphysics)

We can now verify that the courses are there::

  >>> [(course.id, course.name, course.department_id) for course in
  ... session.query(Course)]
  [(1, 'Logic', 1), (2, 'Ethics', 1), (3, 'Metaphysics', 1)]

  >>> for key, value in sorted(philosophy.courses.items()):
  ...   print key, value.name, value.department.name
  1 Logic Philosophy
  2 Ethics Philosophy
  3 Metaphysics Philosophy
"""

from megrok import rdb

from sqlalchemy import Column, ForeignKey
from sqlalchemy.types import Integer, String
from sqlalchemy.orm import relation

metadata = rdb.MetaData()

rdb.metadata(metadata)

class Courses(rdb.Container):
    pass

class Department(rdb.Model):
    id = Column('id', Integer, primary_key=True)
    name = Column('name', String(50))
    courses = relation('Course', 
                       backref='department',
                       collection_class=Courses)

class Course(rdb.Model):
    id = Column('id', Integer, primary_key=True)
    department_id = Column('department_id', Integer, 
                           ForeignKey('department.id'))
    name = Column('name', String(50))




import unittest

import grokcore.component.testing

from megrok import rdb

metadata = rdb.MetaData()
rdb.metadata(metadata)

class Polymorphic(unittest.TestCase):
    def setUp(self):
        grokcore.component.testing.grok('megrok.rdb.meta')
        from megrok.rdb.testing import configureEngine
        engine = configureEngine()

        from sqlalchemy.sql import text
        conn = engine.connect()
        s = text('''
            create table content (
                id integer,
                type char(50),
                primary key (id))
        ''')
        result = conn.execute(s)
        s = text('''
           create table person (
             id integer,
             name char(50),
             FOREIGN KEY(id) REFERENCES content (id)
             )
        ''')
        result = conn.execute(s)
        x = conn.execute
        x('INSERT INTO content VALUES (1, "person")')
        x('INSERT INTO person VALUES (1, "Bob")')
        x('INSERT INTO content VALUES (2, "content")')

    def test_polymorphic_reflection(self):
        """
        Could test basic attributes being set but this integration test
        ensures it continues to work with SQLAlchemy.
        """

        class Content(rdb.Model):
            rdb.reflected()
            rdb.polymorphic_on(table='content', column='type')
            rdb.polymorphic_identity('content')

        class Person(Content):
            rdb.tablename('person')
            rdb.reflected()
            rdb.inherits(Content)
            rdb.polymorphic_identity('person')

        grokcore.component.testing.grok_component('Content', Content)
        grokcore.component.testing.grok_component('Person', Person)

        # this has to be called after classes are grokked
        rdb.setupDatabase(metadata)
        session = rdb.Session()
        
        results = session.query(Content).all()

        assert results
        assert isinstance(results[0], Person)
        # a Person has a name
        assert results[0].name == 'Bob'
        
        assert isinstance(results[1], Content)
        # Content does not have a name
        assert not hasattr(results[1], 'name')
        
    def test_polymorphic_declarative(self):
        """
        Test that non-reflected works
        """
        assert "I am broken!"

def suite():
    return unittest.makeSuite(Polymorphic)




from megrok.rdb.components import Model, Container, QueryContainer
#from megrok.rdb.schema import Fields
from megrok.rdb.directive import (key, metadata, tablename, reflected, table,
                                  tableargs, polymorphic_on, inherits,
                                  polymorphic_identity)
from megrok.rdb.setup import setupDatabase
from megrok.rdb.interfaces import IDatabaseSetupEvent
from megrok.rdb.prop import locatedproperty

from sqlalchemy import MetaData

from z3c.saconfig import Session



import unittest
import doctest
from zope.testing import cleanup
from zope.testing import module
import zope.component.eventtesting
from zope import component
from megrok import rdb
    
from z3c.saconfig.interfaces import IEngineFactory, IScopedSession

from megrok.rdb.tests import tableargs, polymorphic

def moduleSetUp(test):
    # using zope.testing.module.setUp to work around
    # __module__ being '__builtin__' by default
    module.setUp(test, 'foo')
    
def moduleTearDown(test):
    # make sure scope func is empty before we tear down component architecture
    rdb.Session.remove()
    
    module.tearDown(test)
    cleanup.cleanUp()
    
def zopeSetUp(test):
    zope.component.eventtesting.setUp(test)

def zopeTearDown(test):
    rdb.Session.remove()
    cleanup.cleanUp()
    
def test_suite():
    optionflags = doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS
    globs = {}
    
    suite = unittest.TestSuite()
    
    suite.addTest(doctest.DocFileSuite(
        '../README.txt',
        optionflags=optionflags,
        setUp=moduleSetUp,
        tearDown=moduleTearDown,
        globs=globs))
    suite.addTest(doctest.DocTestSuite(
        'megrok.rdb.tests.creation',
        setUp=zopeSetUp,
        tearDown=zopeTearDown,
        optionflags=optionflags))
    suite.addTest(doctest.DocTestSuite(
        'megrok.rdb.tests.explicittable',
        setUp=zopeSetUp,
        tearDown=zopeTearDown,
        optionflags=optionflags))
    suite.addTest(doctest.DocTestSuite(
        'megrok.rdb.tests.reflection',
        setUp=zopeSetUp,
        tearDown=zopeTearDown,
        optionflags=optionflags))
    suite.addTest(doctest.DocTestSuite(
        'megrok.rdb.tests.location',
        setUp=zopeSetUp,
        tearDown=zopeTearDown,
        optionflags=optionflags))
    suite.addTest(doctest.DocFileSuite(
        '../schema.txt',
        optionflags=optionflags,
        ))
    suite.addTest(tableargs.suite())
    suite.addTest(polymorphic.suite())
    return suite



from setuptools import setup, find_packages
import os

def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()

long_description = read('src/megrok/rdb/README.txt') + '\n' +\
    read('CHANGES.txt')

setup(name='megrok.rdb',
      version = '0.9',
      description="SQLAlchemy based RDB support for Grok.",
      long_description=long_description,
      classifiers=['Development Status :: 4 - Beta',
                   'Environment :: Web Environment',
                   'Framework :: Zope3',
                   'Intended Audience :: Developers',
                   'License :: OSI Approved :: Zope Public License',
                   'Operating System :: OS Independent',
                   'Programming Language :: Python',
                   'Topic :: Internet :: WWW/HTTP',
                   'Topic :: Software Development :: Libraries',
                   'Topic :: Database',
                   ],
      keywords='rdb relational sqlalchemy grok database',
      author='Grok Team',
      author_email='grok-dev@zope.org',
      license='ZPL',
      packages=find_packages('src'),
      package_dir = {'': 'src'},
      namespace_packages=['megrok'],
      include_package_data=True,
      zip_safe=True,
      install_requires=[
          'setuptools',
          'zope.interface',
          'zope.schema',
          'zope.component',
          'zope.event',
          'zope.location',
          'grokcore.component',
          'zope.app.container', # want to depend on zope.container directly but can't yet to retain grok compatibility
          'martian < 0.12', # 0.12 is too new for grokcore.component right now..
          'SQLAlchemy > 0.5beta2',
          'zope.sqlalchemy',
          'z3c.saconfig >= 0.9',
         ],
      entry_points="""
      # -*- Entry points: -*-
      """,
      )



# faasen's proof of concept of a sorting by inheritance
class C(object):
    pass

class B(C):
    pass

class A(B):
    pass

d = {
    A: B,
    B: C,
    C: None,
    }

l = [A, B, C]

def sorted_inheritance(x, y):
    if x is y:
        return 0
    if inherits(x, y):
        return 1
    else:
        return -1

def inherits(x, y):
    if x is y:
        return True
    potential_base = d[x]
    if potential_base is None:
        return False
    if potential_base is y:
        return True
    return inherits(potential_base, y)

print sorted(l, cmp=sorted_inheritance)



from zope import component
from zope.event import notify

from sqlalchemy.ext.declarative import instrument_declarative
from sqlalchemy.schema import Table

from z3c.saconfig.interfaces import IEngineFactory
from megrok.rdb.interfaces import DatabaseSetupEvent

def setupDatabase(metadata):
    """Set up of ORM for engine in current site.

    This will:

    * reflect any reflected tables that need to be reflected from the database
      into classes.

    * create any tables in the database that haven't been yet reflected.    
    """
    reflectTables(metadata)
    createTables(metadata)
    notify(DatabaseSetupEvent(metadata))
    
def reflectTables(metadata):
    """Reflect tables into ORM.
    """
    if getattr(metadata, '_reflected_completed', False):
        # XXX thread safety?
        return
    if not getattr(metadata, '_reflected_registry', {}):
        # nothing to reflect
        return
    # first reflect database-defined schemas into metadata
    engine = Engine()
    for class_ in metadata._reflected_registry.keys():
        _reflectTableForClass(class_, metadata, engine)
    # reflect any remaining tables. This will not reload tables already loaded
    # (XXX is this necessary?)
    metadata.reflect(bind=engine)
    if not hasattr(metadata, '_decl_registry'):
        metadata._decl_registry = {}

    # XXX should sort by the inheritance tree
    for class_ in sorted(metadata._reflected_registry.keys(), key=lambda a: getattr(a, 'megrok.rdb.directive.inherits', 0)):
        class_args = metadata._reflected_registry[class_]
        polymorphic_on = class_args['polymorphic_on']
        polymorphic_identity = class_args['polymorphic_identity']
        inherits = class_args['inherits']
        mapper_args = getattr(class_, '__mapper_args__', {})

        if polymorphic_on:
            tablename, column = polymorphic_on
            mapper_args['polymorphic_on'] = getattr(metadata.tables[tablename].c, column)
        if polymorphic_identity:
            mapper_args['polymorphic_identity'] = polymorphic_identity
        if inherits:
            mapper_args['inherits'] = inherits
                    
        if mapper_args:
            class_.__mapper_args__ = mapper_args

        instrument_declarative(class_, metadata._decl_registry, metadata)

    # XXX thread safety?
    metadata._reflected_completed = True

def _reflectTableForClass(class_, metadata, bind):
    """Reflect an individual table defined for a class.

    This supports the special 'schema' indicator in '__table_args__',
    allowing reflection of tables not in the default schema.
    metadata.reflect() doesn't allow this in declarative mode.
    """
    reflect_opts = {'autoload': True}
    reflect_opts['autoload_with'] = bind
    conn = bind.contextual_connect()

    table_args = getattr(class_, '__table_args__', {})
    if type(table_args) is tuple:
        table_args = table_args[-1]
        if type(table_args) is not dict:
            table_args = {}
    schema = table_args.get('schema', None)
    if schema is not None:
        reflect_opts['schema'] = schema

    Table(class_.__tablename__, metadata, **reflect_opts)
        
def createTables(metadata):
    """Create class-specified tables.
    """
    engine = Engine()
    metadata.create_all(engine)

def Engine():
    """Get the engine in the current session.
    """
    engine_factory = component.getUtility(IEngineFactory)
    return engine_factory()
